/*
 * tcpmon.c - Linux 2.2 kernel TCP monitor (part of tcpspy)
 * Copyright (c) 2000, 2001 Tim J. Robbins <tim@robbins.dropbear.id.au>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Brief overview:
 * 	Creates a file /proc/tcpmon from which messages are read by a
 * 	userspace process. These messages are written when a connect(),
 * 	listen() or accept() call occurs, and contain information about
 * 	the current process, and the local and remote network addresses
 * 	(if applicable).
 *
 * To do:
 *	- Ensure that this module is SMP-safe, particularly the replacement
 *	socketcall.
 *	- Find a way to log when a socket is closed - replacing the close
 *	syscall doesn't work because sockets can be closed in other ways.
 *
 * Contains code from various places in the Linux kernel source:
 * 	linux/kernel/printk.c	Linus Torvalds
 * 	linux/net/socket.c	Orest Zborowski, Ross Biro, Fred N. van Kempen
 * 	linux/fs/proc/kmsg.c	Linus Torvalds
 *
 * $Id: tcpmon.c,v 1.1 2001/06/19 13:03:43 tim Stab $
 */

#include <linux/config.h>
#include <linux/module.h>
/*#include <linux/modversions.h>*/
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <linux/poll.h>
#include <linux/sched.h>
#include <linux/net.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/in.h>
#include <linux/file.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <net/sock.h>
#include <sys/syscall.h>

/* Prototypes */
void cleanup_module (void);
int init_module (void);
static struct socket *tcpmon_sockfd_lookup (int fd, int *err);
static void tcpmon_sockfd_put(struct socket *sock);
static struct socket *tcpmon_socki_lookup(struct inode *inode);
asmlinkage int tcpmon_sys_close (unsigned int fd);
asmlinkage long tcpmon_sys_socketcall (int call, unsigned long *args);
static int tcpmon_do_log(int type, char * buf, int len);
static int tcpmon_printl (const char *fmt, ...);
static ssize_t tcpmon_read (struct file * file, char * buf, size_t count, loff_t *ppos);
static unsigned int tcpmon_poll(struct file *file, poll_table * wait);

/* Data */
#define PROC_TCPMON 4095 /* XXX */
extern void *sys_call_table[];
static long (*old_sys_socketcall) (int, unsigned long *);
#define TLOG_BUF_LEN	(16384)
#define TLOG_BUF_MASK	(TLOG_BUF_LEN-1)
static char buf[1024];
static unsigned long tlog_size = 0;
static struct wait_queue * tlog_wait = NULL;
spinlock_t tlog_lock = SPIN_LOCK_UNLOCKED;
static char tlog_buf[TLOG_BUF_LEN];
static unsigned long tlog_start = 0;
static unsigned long tlogged_chars = 0;
struct file_operations proc_tcpmon_operations;
struct proc_dir_entry proc_root_tcpmon;
struct inode_operations proc_tcpmon_inode_operations;

/*
 * Part 1: General module functions.
 */

int init_module (void)
{
	proc_register(&proc_root, &proc_root_tcpmon); /* XXX check errors */

	old_sys_socketcall = sys_call_table[__NR_socketcall];
	sys_call_table[__NR_socketcall] = tcpmon_sys_socketcall;

	printk (KERN_INFO "tcpmon loaded\n");
	
	return 0;
}


void cleanup_module (void)
{
	if (sys_call_table[__NR_socketcall] != tcpmon_sys_socketcall)
		printk (KERN_ERR "tcpmon: syscall table modified\n");
	sys_call_table[__NR_socketcall] = old_sys_socketcall;

	proc_unregister(&proc_root, proc_root_tcpmon.low_ino);
	
	printk (KERN_INFO "tcpmon unloaded\n");
}


/*
 * Part 2: Replacement syscalls.
 */

/* Replacement for the socketcall syscall that logs information */
asmlinkage long tcpmon_sys_socketcall (int call, unsigned long *args)
{
	int ret, err;
	struct socket *sock;
	
	if ((ret = old_sys_socketcall (call, args)) < 0)
		return ret;

	/*lock_kernel ();*/

	if (call == SYS_CONNECT) {
		int fd;

		if (copy_from_user (&fd, args, sizeof (*args)) < 0)
			goto out;

		if ((sock = tcpmon_sockfd_lookup (fd, &err)) == NULL)
			goto out;

		if ((sock->sk->family != PF_INET) || (sock->sk->protocol != IPPROTO_TCP))
			goto out_connect;

		tcpmon_printl ("CONNECT %s %d %d %08x %04x %08x %04x\n",
				current->comm, current->pid, current->uid,
				sock->sk->daddr, sock->sk->dport,
				sock->sk->rcv_saddr, sock->sk->num
				);
out_connect:
		tcpmon_sockfd_put (sock);

	} else if (call == SYS_LISTEN) {
		int fd;
		
                if (copy_from_user (&fd, args, sizeof (*args)) != 0)
			goto out;
		
		if ((sock = tcpmon_sockfd_lookup (fd, &err)) == NULL)
			goto out;

		if ((sock->sk->family != PF_INET) || (sock->sk->protocol != IPPROTO_TCP))
			goto out_listen;

		tcpmon_printl ("LISTEN %s %d %d %08x %04x\n",
				current->comm, current->pid, current->uid,
				sock->sk->rcv_saddr, sock->sk->num);

out_listen:
		tcpmon_sockfd_put (sock);

	} else if (call == SYS_ACCEPT) {
		int fd;

		if (copy_from_user (&fd, args, sizeof (*args)) != 0)
			goto out;

		if ((sock = tcpmon_sockfd_lookup (ret, &err)) == NULL)
			goto out;

		if ((sock->sk->family != PF_INET) || (sock->sk->protocol != IPPROTO_TCP))
			goto out_accept;

		tcpmon_printl ("ACCEPT %s %d %d %08x %04x %08x %04x\n",
				current->comm, current->pid, current->uid,
				sock->sk->daddr, sock->sk->dport,
				sock->sk->rcv_saddr, sock->sk->num);

out_accept:
		tcpmon_sockfd_put (sock);
	}

out:
	/*unlock_kernel ();*/

	return ret;
}


static struct socket *tcpmon_sockfd_lookup (int fd, int *err)
{
	struct file *file;
	struct inode *inode;
	struct socket *sock;

	if (!(file = fget(fd)))
	{
		*err = -EBADF;
		return NULL;
	}

	inode = file->f_dentry->d_inode;
	if (!inode || !inode->i_sock || !(sock = tcpmon_socki_lookup(inode)))
	{
		*err = -ENOTSOCK;
		fput(file);
		return NULL;
	}

	if (sock->file != file) {
		printk(KERN_ERR "tcpmon_sockfd_lookup: socket file changed!\n");
		sock->file = file;
	}
	return sock;
}


static void tcpmon_sockfd_put(struct socket *sock)
{
	fput(sock->file);
}


static struct socket *tcpmon_socki_lookup(struct inode *inode)
{
	return &inode->u.socket_i;
}


/*
 * Part 3: /proc/tcpmon log file.
 */

struct file_operations proc_tcpmon_operations = {
        NULL,
        tcpmon_read,
        NULL,
        NULL,
        tcpmon_poll,
        NULL,
        NULL,
	NULL,
        NULL,
	NULL,
        NULL
};

struct inode_operations proc_tcpmon_inode_operations = {
        &proc_tcpmon_operations,/* default base directory file-ops */
        NULL,                   /* create */
        NULL,                   /* lookup */
        NULL,                   /* link */
        NULL,                   /* unlink */
        NULL,                   /* symlink */
        NULL,                   /* mkdir */
        NULL,                   /* rmdir */
        NULL,                   /* mknod */
        NULL,                   /* rename */
        NULL,                   /* readlink */
        NULL,                   /* follow_link */
        NULL,                   /* readpage */
        NULL,                   /* writepage */
        NULL,                   /* bmap */
        NULL,                   /* truncate */
        NULL                    /* permission */
};

struct proc_dir_entry proc_root_tcpmon = {
        PROC_TCPMON, 6, "tcpmon",
        S_IFREG | S_IRUSR, 1, 0, 0,
        0, &proc_tcpmon_inode_operations
};


ssize_t tcpmon_read (struct file * file, char * buf, size_t count, loff_t *ppos)
{
	return tcpmon_do_log(2,buf,count);
}


unsigned int tcpmon_poll(struct file *file, poll_table * wait)
{
	poll_wait(file, &tlog_wait, wait);
	if (tlog_size)
		return POLLIN | POLLRDNORM;
	return 0;
}


/* Perform operations on the log buffer, same arguments as do_syslog() */
static int tcpmon_do_log(int type, char * buf, int len)
{
	unsigned long i, j, limit, count;
	int do_clear = 0;
	char c;
	int error = -EPERM;

	error = 0;
	switch (type) {
	case 0:		/* Close log */
		break;
	case 1:		/* Open log */
		break;
	case 2:		/* Read from log */
		error = -EINVAL;
		if (!buf || len < 0)
			goto out;
		error = 0;
		if (!len)
			goto out;
		error = verify_area(VERIFY_WRITE,buf,len);
		if (error)
			goto out;
		error = wait_event_interruptible(tlog_wait, tlog_size);
		if (error)
			goto out;
		i = 0;
		spin_lock_irq(&tlog_lock);
		while (tlog_size && i < len) {
			c = tlog_buf[tlog_start & TLOG_BUF_MASK];
			tlog_start++;
			tlog_size--;
			spin_unlock_irq(&tlog_lock);
			__put_user(c,buf);
			buf++;
			i++;
			spin_lock_irq(&tlog_lock);
		}
		spin_unlock_irq(&tlog_lock);
		error = i;
		break;
	case 4:		/* Read/clear last kernel messages */
		do_clear = 1; 
		/* FALL THRU */
	case 3:		/* Read last kernel messages */
		error = -EINVAL;
		if (!buf || len < 0)
			goto out;
		error = 0;
		if (!len)
			goto out;
		error = verify_area(VERIFY_WRITE,buf,len);
		if (error)
			goto out;
		spin_lock_irq(&tlog_lock);
		count = len;
		if (count > TLOG_BUF_LEN)
			count = TLOG_BUF_LEN;
		if (count > tlogged_chars)
			count = tlogged_chars;
		if (do_clear)
			tlogged_chars = 0;
		limit = tlog_start + tlog_size;
		/*
		 * __put_user() could sleep, and while we sleep
		 * printk() could overwrite the messages 
		 * we try to copy to user space. Therefore
		 * the messages are copied in reverse. <manfreds>
		 */
		for(i=0;i < count;i++) {
			j = limit-1-i;
			if (j+TLOG_BUF_LEN < tlog_start+tlog_size)
				break;
			c = tlog_buf[ j  & TLOG_BUF_MASK ];
			spin_unlock_irq(&tlog_lock);
			__put_user(c,&buf[count-1-i]);
			spin_lock_irq(&tlog_lock);
		}
		spin_unlock_irq(&tlog_lock);
		error = i;
		if(i != count) {
			int offset = count-error;
			/* buffer overflow during copy, correct user buffer. */
			for(i=0;i<error;i++) {
				__get_user(c,&buf[i+offset]);
				__put_user(c,&buf[i]);
			}
		}

		break;
	case 5:		/* Clear ring buffer */
		spin_lock_irq(&tlog_lock);
		tlogged_chars = 0;
		spin_unlock_irq(&tlog_lock);
		break;
	default:
		error = -EINVAL;
		break;
	}
out:
	return error;
}


/* Write a log message to be read through /proc/tcpmon, similar to printk */
static int tcpmon_printl (const char *fmt, ...)
{
	va_list args;
	int i;
	char *msg, *p, *buf_end;
	int line_feed;
	static signed char msg_level = -1;
	long flags;

	spin_lock_irqsave(&tlog_lock, flags);
	va_start(args, fmt);
	i = vsprintf(buf, fmt, args); /* hopefully i < sizeof(buf)-4 */
	buf_end = buf + i;
	va_end(args);
	for (p = buf; p < buf_end; p++) {
		msg = p;
		line_feed = 0;
		for (; p < buf_end; p++) {
			tlog_buf[(tlog_start+tlog_size) & TLOG_BUF_MASK] = *p;
			if (tlog_size < TLOG_BUF_LEN)
				tlog_size++;
			else
				tlog_start++;

			tlogged_chars++;
			if (*p == '\n') {
				line_feed = 1;
				break;
			}
		}
		if (line_feed)
			msg_level = -1;
	}
	spin_unlock_irqrestore(&tlog_lock, flags);
	wake_up_interruptible(&tlog_wait);
	return i;
}

