/*
 * tcpspy.c - TCP/IP connection monitor.
 *
 * This file is part of tcpspy, a TCP/IP connection monitor.
 *
 * Copyright (c) 2000, 2001 Tim J. Robbins. 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: tcpspy.c,v 2.2 2001/06/19 13:23:53 tim Exp $
 */

#include <arpa/inet.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <limits.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define SYSLOG_NAMES
#include <syslog.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "log.h"
#include "rcsid.h"
#include "rule.h"

RCSID("$Id: tcpspy.c,v 2.2 2001/06/19 13:23:53 tim Exp $");

/*
 * Defaults for compile-time settings. Descriptions of these are in
 * the Makefile.
 */
#ifndef FACILITY
#define FACILITY LOG_USER
#endif

static int gotrule = 0;
static int facility = FACILITY;


/*
 * getfacility ()
 *
 * Convert a facility name to a integer facility value for syslog().
 */
int getfacility (const char *s)
{
	int i;

	assert (s != NULL);

	for (i = 0; i < LOG_NFACILITIES - 2; i++) {
		assert (facilitynames[i].c_name != NULL);
		if (strcmp (facilitynames[i].c_name, s) == 0)
			return facilitynames[i].c_val;
	}

	return -1;
}

static void portlookup (unsigned short port, char *namebuf, size_t nbsiz)
{
	struct servent *se;

	assert ((namebuf != NULL) && (nbsiz > 0));

	if ((se = getservbyport (htons (port), "tcp")) != NULL) {
		strncpy (namebuf, se->s_name, nbsiz);
		namebuf[nbsiz - 1] = '\0';
	} else {
		snprintf (namebuf, nbsiz, "%u", port);
	}
}

static void userlookup (int uid, char *userbuf, size_t ubsiz)
{
	struct passwd *pw;

	assert ((userbuf != NULL) && (ubsiz > 0));

	if ((pw = getpwuid ((uid_t) uid)) != NULL) {
		strncpy (userbuf, pw->pw_name, ubsiz);
		userbuf[ubsiz - 1] = '\0';
	} else {
		snprintf (userbuf, ubsiz, "%u", (unsigned) uid);
	}
}

static void process_line (const char *line)
{
	char *params;

	assert (line != NULL);

	if ((params = strchr (line, ' ')) != NULL)
		params++; /* skip over the space */

	if ((strncmp (line, "CONNECT", strlen ("CONNECT")) == 0) ||
		(strncmp (line, "ACCEPT", strlen ("ACCEPT")) == 0)) {
		char comm[256];
		int pid, uid;
		unsigned long raddr, laddr;
		unsigned short rport, lport;
		int n;

		if (params == NULL)
			panic ("/proc/tcpmon: missing parameters for CONNECT/ACCEPT");

		n = sscanf (params, "%255s %d %d %lx %hx %lx %hx",
				comm, &pid, &uid, &raddr, &rport,
				&laddr, &lport);
		if (n != 7)
			panic ("/proc/tcpmon: bad parameters for CONNECT/ACCEPT");

		if ((gotrule == 0) || (rule_eval (uid, laddr, lport, raddr, rport, comm) != 0)) {
			char traddr[16], tladdr[16];
			char trport[32], tlport[32];
			char username[256];
			struct in_addr in;
			char *n;
			const char *arrow;

			in.s_addr = laddr;
			n = inet_ntoa (in);
			strncpy (tladdr, n, sizeof (tladdr));
			tladdr[sizeof(tladdr)-1] = '\0';
			in.s_addr = raddr;
			n = inet_ntoa (in);
			strncpy (traddr, n, sizeof (traddr));
			tladdr[sizeof(traddr)-1] = '\0';

			arrow = (*line == 'C') ? "->" : "<-";

			portlookup (rport, trport, sizeof (trport));
			portlookup (lport, tlport, sizeof (tlport));
			userlookup (uid, username, sizeof (username));

			logmsg ("%s[%d] %s(%d) %s:%s %s %s:%s",
					comm, pid, username, uid, tladdr,
					tlport, arrow, traddr, trport);

		}
	}
}

static void usage (void)
{
	fprintf (stderr, "usage: tcpspy [-dp] [-e rule]... [-f rulefile] [-I interval] "
			"[-U user] [-G group]\n\t[-F facility] [-I interval]\n");
	exit (EXIT_FAILURE);
}

int main (int argc, char *argv[])
{
	int ch;
	uid_t dropuser = -1;
	gid_t dropgroup = -1, defgroup = -1;
	int debug = 0, firstrule = 1;
	int monfd;
	FILE *monfile;
	char linebuf[1024];

	log_set_syslog ();

	/*
	 * Parse our arguments.
	 */
	opterr = 0;
	while ((ch = getopt (argc, argv, "dpe:I:U:G:u:w:i:f:F:")) != -1) {
		switch (ch) {
			case 'd':
				debug = 1;
				log_set_stdout ();
				break;
			case 'e':
				rule_parse (optarg);
				gotrule = 1;
				if (firstrule == 0) 
					/*
					 * Join multiple rules together with
					 * logical OR's.
					 */
					rule_gen_or ();
				firstrule = 0;
				break;
			case 'f':
			{
				FILE *fp;
				char fbuf[512];
				unsigned lineno = 0;
				char *c;

				if ((fp = fopen (optarg, "r")) == NULL) {
					fprintf (stderr, "tcpspy: %s: %s\n", optarg, strerror (errno));
					exit (EXIT_FAILURE);
				}

				while (fgets (fbuf, sizeof fbuf, fp) != NULL) {
					lineno++;
					if (fbuf[strlen (fbuf) - 1] != '\n') {
						fprintf (stderr, "tcpspy: %s:%u: line too long or missing newline\n", optarg, lineno);
						exit (EXIT_FAILURE);
					}
					/*
					 * Remove newline, remove comments.
					 */
					fbuf[strlen (fbuf) - 1] = '\0';
					if ((c = strchr (fbuf, '#')) != NULL)
						*c = '\0';
					/*
					 * Skip blank lines or whole line comments.
					 */
					if (*fbuf == '\0') continue;

					rule_parse (fbuf);
					if (firstrule == 0) rule_gen_or ();
					gotrule = 1;
					firstrule = 0;
				}

				fclose (fp);
			}
			break;
			case 'U':
				{
				struct passwd *pw;
				
				if (isdigit (*optarg)) {
					dropuser = atoi (optarg);
					pw = getpwuid (atoi (optarg));
				} else {
					if ((pw = getpwnam (optarg)) == NULL) {
						fprintf (stderr, "tcpspy: user `%s' unknown\n", optarg);
						exit (EXIT_FAILURE);
					}
					dropuser = pw->pw_uid;
				}
				
				/*
				 * Use the gid from the password file entry if
				 * possible, as a default.
				 */
				if (pw != NULL)
					defgroup = pw->pw_gid;
				else
					defgroup = (gid_t) -1;
				
				}
				break;
			case 'G':
				if (isdigit (*optarg))
					dropgroup = atoi (optarg);
				else {
					struct group *gr;

					if ((gr = getgrnam (optarg)) == NULL) {
						fprintf (stderr, "tcpspy: group `%s' unknown\n", optarg);
						exit (EXIT_FAILURE);
					}
					dropgroup = gr->gr_gid;
				}	
				break;
			case 'F':
				if ((facility = getfacility (optarg)) < 0) {
					fprintf (stderr, "tcpspy: bad facility name `%s'\n", optarg);
					exit (EXIT_FAILURE);
				}
				break;
			case 'u': case 'w': case 'i': case 'p': case 'I':
				fprintf (stderr, "tcpspy: -%c option is obsolete\n", ch);
				/* fall through to usage message */
			default:
				usage();
		}
	}

	argc -= optind;
	argv += optind;

	/*
	 * Open /proc/tcpmon before we drop privileges
	 */
	if ((monfd = open ("/proc/tcpmon", O_RDONLY)) < 0) {
		fprintf (stderr, "tcpspy: /proc/tcpmon: %s\n", strerror (errno));
		exit (EXIT_FAILURE);
	}

	/*
	 * Become an unprivileged user for safety purposes if requested.
	 */
	if ((dropgroup == (uid_t) -1) && (defgroup != (uid_t) -1))
		dropgroup = defgroup;
	if (dropgroup != (gid_t) -1) {
		if (setgid (dropgroup) < 0) {
			fprintf (stderr, "tcpspy: setgid: %s\n", strerror (errno));
			exit (EXIT_FAILURE);
		}
	}
	if (dropuser != (uid_t) -1) {
		if (setuid (dropuser) < 0) {
			fprintf (stderr, "tcpspy: setuid: %s\n", strerror (errno));
			exit (EXIT_FAILURE);
		}
	}

	/*
	 * Become a daemon..
	 */

	if (debug == 0) {
		pid_t p;
		p = fork();
		if (p < 0) {
			fprintf (stderr, "tcpspy: fork: %s\n",
					strerror (errno));
			exit (EXIT_FAILURE);
		} else if (p != 0) {
			fprintf (stderr, "tcpspy 2.0-pre started (pid %d)\n", 
					(int) p);
			exit (EXIT_SUCCESS);
		}

		ioctl (STDIN_FILENO, TIOCNOTTY, NULL);
		close (STDIN_FILENO); 
		close (STDOUT_FILENO); 
		close (STDERR_FILENO);
	} else
		fprintf (stderr, "tcpspy 2.0-pre started (debug)\n");


	/*
	 * Initialisation's done, start watching for connections.
	 */

	openlog ("tcpspy", LOG_PID, facility);

	if ((monfile = fdopen (monfd, "r")) == NULL)
		panic ("fdopen tcpmon descriptor: %s", strerror (errno));

	while (fgets (linebuf, sizeof (linebuf), monfile) != NULL) {
		if (linebuf[strlen(linebuf)-1] != '\n')
			panic ("/proc/tcpmon: line too long");
		linebuf[strlen(linebuf)-1] = '\0';
		process_line (linebuf);
	}

	closelog ();

	return EXIT_SUCCESS;
}
